"""
    scan_xy(scanner, x_axis_range, x_axis_num_scans, y_axis_range, y_axis_num_scans; kwargs..)
    scan_xz(scanner, x_axis_range, x_axis_num_scans, z_axis_range, z_axis_num_scans; kwargs..)
    scan_yz(scanner, y_axis_range, y_axis_num_scans, z_axis_range, z_axis_num_scans; kwargs..)

This function grabs data along the xy-plane. It holds the y-position
constant and moves along the x axis until it reaches the last x point 
and then changes the y-position and repeats. There are five mandatory 
inputs and one optional input that is used only when this function is
called by other functions and the printed messages should be 
surpressed. The user usually doesn't have to use the variable called.

# Arguments
- `scanner`: the structure for all controller connections.
- `x_axis_range::Array`: An array that contains the start and end point
                         of the desired range in the x-direction.
                         (Must be 1 or 2 elements long)
- `x_axis_num_scans::Int`: The number of points/scans the user desires to
                           record along the x-axis.
- `y_axis_range::Array`: Contains the start and end point
                         of the desired range in the
                         y-direction.
                         (Must be 1 or 2 elements long)
- `y_axis_num_scans::Int`: The number of points/scans the user desires 
                           to record along the y-axis.

# Keywords
- `verbose`::Bool: Optional, defaults to: true

# Returns
- `ScanData`

# Example
```
ps = initialize(PositionerSystem)
scope = initialize(AgilentDSOX4034A)

scanner = IntensityScan(
    xyz = ps, 
    scope = scope, 
    channel = 1, 
    post_move_delay = 0ms,
    precapture_delay = 0µs,
)

wave_xy = scan_xy(scanner, [-5mm, 5mm], 3, [-5mm, 5mm], 7)
```
"""
function scan_xy(scanner, x_axis_range, x_axis_num_scans, y_axis_range, y_axis_num_scans; verbose=true)
    scan_data = scan_double_axis(
        scanner,
        x_axis_range,
        x_axis_num_scans,
        scan_x,
        y_axis_range,
        y_axis_num_scans,
        move_y_abs;
        verbose=verbose,
    )
    return scan_data
end

function scan_xz(scanner, x_axis_range, x_axis_num_scans, z_axis_range, z_axis_num_scans; verbose=true)
    scan_data = scan_double_axis(
        scanner,
        x_axis_range,
        x_axis_num_scans,
        scan_x,
        z_axis_range,
        z_axis_num_scans,
        move_z_abs;
        verbose=verbose,
    )
    return scan_data
end

function scan_yz(scanner, y_axis_range, y_axis_num_scans, z_axis_range, z_axis_num_scans; verbose=true)
    scan_data = scan_double_axis(
        scanner,
        y_axis_range,
        y_axis_num_scans,
        scan_y,
        z_axis_range,
        z_axis_num_scans,
        move_z_abs;
        verbose=verbose,
    )
    return scan_data
end


function scan_double_axis(
    scanner::IntensityScan,
    first_axis_range,
    first_axis_num_scans,
    scan_first_axis::Function,
    second_axis_range,
    second_axis_num_scans,
    move_second_axis::Function;
    verbose=true,
)
    first_axis = get_axis(scan_first_axis)
    second_axis = get_axis(move_second_axis)
    axes = "$first_axis$second_axis"
    check_xyz_limits(scanner, second_axis, second_axis_range)
    second_axis_positions = create_positions_vector(second_axis_range, second_axis_num_scans)
    
    # construct the ScanData struct
    scope_info = TcpInstruments.get_waveform_info(scanner.scope, scanner.channel)
    dummy_data = TcpInstruments.parse_raw_waveform(zeros(scope_info.num_points), scope_info)
    time_points = dummy_data.time
    num_points_per_axis = [first_axis_num_scans, second_axis_num_scans]
    scan_data = ScanData(axes, scope_info, time_points, scope_info.num_points, num_points_per_axis)
    
    pm = Progress(second_axis_num_scans; desc="Scanning $axes-plane: ", enabled=verbose)
    for scan_index in 1:second_axis_num_scans
        move_second_axis(scanner.xyz, round(u"mm", second_axis_positions[scan_index]; digits=2))
        ProgressMeter.update!(pm)

        scan_data_first_axis = scan_first_axis(scanner, first_axis_range, first_axis_num_scans; verbose=false)
        scan_data.waveform[:, :, scan_index] = scan_data_first_axis.waveform
        scan_data.coordinates[:, :, scan_index] = scan_data_first_axis.coordinates

        ProgressMeter.next!(pm)
    end

    return scan_data
end
