# 1D data
@recipe function f(metric::Metric{T, 1}; axes="nil", coordinates=nothing) where {T <: UltrasoundMetric}
    axes == "nil" && error("Keyword missing: axes=\"x\" or axes=\"y\"..")
    check_coordinates(coordinates)

    title --> get_metric_display_name(metric)
    xguide --> "$(axes)-axis"
    legend --> false

    target_dimension = 1
    axis_coordinates = get_axis_coordinates(axes, target_dimension, coordinates)

    return axis_coordinates, metric.val
end

function check_coordinates(coordinates)
    if isnothing(coordinates)
        error("You must set coordinates from scan in keyword arg:
              ;coordinates=scan.coordinates")
    end
end

function get_axis_coordinates(axis_name, target_dimension, coordinates)
    axis_name = Val(Symbol(axis_name))
    target_dimension = Val(target_dimension)
    axis_coordinates = get_coordinates(axis_name, get_dimension(target_dimension, coordinates))
    return round.(typeof(1.0mm), uconvert.(mm, axis_coordinates); digits=2)
end

get_dimension(::Val{1}, coordinates) = coordinates[:, :, 1, 1]
get_dimension(::Val{2}, coordinates) = coordinates[:, 1, :, 1]
get_dimension(::Val{3}, coordinates) = coordinates[:, 1, 1, :]

get_coordinates(::Val{:x}, coordinates) = coordinates[1, :]
get_coordinates(::Val{:y}, coordinates) = coordinates[2, :]
get_coordinates(::Val{:z}, coordinates) = coordinates[3, :]


# 2D data
@recipe function f(metric::Metric{T, 2}; axes="nil", coordinates=nothing) where {T <: UltrasoundMetric}
    axes == "nil" && error("Keyword missing: axes=\"xy\" or axes=\"yz\"..")
    check_coordinates(coordinates)

    seriestype := :heatmap
    title --> get_metric_display_name(metric)
    xguide --> "$(axes[1])-axis"
    yguide --> "$(axes[2])-axis"
    aspect_ratio --> :equal

    metric_values_2d = metric.val
    
    axis_1_coordinates = get_axis_coordinates(axes[1], 1, coordinates)
    axis_2_coordinates = get_axis_coordinates(axes[2], 2, coordinates)
    axis_1_step_size = diff(axis_1_coordinates)[1]
    axis_2_step_size = diff(axis_2_coordinates)[1]
    if axis_1_step_size < 0mm
        axis_1_coordinates = reverse(axis_1_coordinates)
        axis_1_step_size = diff(axis_1_coordinates)[1]
        metric_values_2d = reverse(metric_values_2d, dims=1)
    end
    if axis_2_step_size < 0mm
        axis_2_coordinates = reverse(axis_2_coordinates)
        axis_2_step_size = diff(axis_2_coordinates)[1]
        metric_values_2d = reverse(metric_values_2d, dims=2)
    end
    xlims --> (axis_1_coordinates[1]-axis_1_step_size/2, axis_1_coordinates[end]+axis_1_step_size/2)
    ylims --> (axis_2_coordinates[1]-axis_2_step_size/2, axis_2_coordinates[end]+axis_2_step_size/2)

    return axis_1_coordinates, axis_2_coordinates, transpose(metric_values_2d)
end

# 3D data
@recipe function f(metric::Metric{T, 3}; coordinates=nothing, axes="xyz") where {T <: UltrasoundMetric}
    check_coordinates(coordinates)

    # remove the keywords from the attribute dictionary to avoid potential errors
    xslice = pop!(plotattributes, :xslice)
    yslice = pop!(plotattributes, :yslice)
    zslice = pop!(plotattributes, :zslice)

    slice_position, slice_axis_name, slice_axis_dim = get_slice_info(xslice, yslice, zslice)
    slice_axis_coordinates = get_axis_coordinates(slice_axis_name, slice_axis_dim, coordinates)
    slice_idx, nearest_slice_position = get_slice_idx(slice_axis_coordinates, slice_position)
    metric_values_2d = get_slice_data(Val(slice_axis_name), slice_idx, metric.val)
    _, slice_image_dims = get_slice_image_axes(Val(slice_axis_name))

    seriestype := :heatmap
    title := get_metric_display_name(metric) * "\n($(slice_axis_name)=$nearest_slice_position)"
    xguide --> "$(axes[slice_image_dims[1]])-axis"
    yguide --> "$(axes[slice_image_dims[2]])-axis"
    aspect_ratio --> :equal
    
    axis_1_coordinates = get_axis_coordinates(axes[slice_image_dims[1]], slice_image_dims[1], coordinates)
    axis_2_coordinates = get_axis_coordinates(axes[slice_image_dims[2]], slice_image_dims[2], coordinates)
    axis_1_step_size = diff(axis_1_coordinates)[1]
    axis_2_step_size = diff(axis_2_coordinates)[1]
    if axis_1_step_size < 0mm
        axis_1_coordinates = reverse(axis_1_coordinates)
        axis_1_step_size = diff(axis_1_coordinates)[1]
        metric_values_2d = reverse(metric_values_2d, dims=1)
    end
    if axis_2_step_size < 0mm
        axis_2_coordinates = reverse(axis_2_coordinates)
        axis_2_step_size = diff(axis_2_coordinates)[1]
        metric_values_2d = reverse(metric_values_2d, dims=2)
    end
    xlims --> (axis_1_coordinates[1]-axis_1_step_size/2, axis_1_coordinates[end]+axis_1_step_size/2)
    ylims --> (axis_2_coordinates[1]-axis_2_step_size/2, axis_2_coordinates[end]+axis_2_step_size/2)

    return axis_1_coordinates, axis_2_coordinates, transpose(metric_values_2d)
end

function get_slice_info(xslice, yslice, zslice)
    if reduce(+, isnan.([xslice, yslice, zslice])) != 2
        error("Only one of the keyword arguments must be set:
              xslice, yslice, or zslice")
    end
    if !isnan(xslice) 
        slice_position = xslice
        slice_axis_name = :x
        slice_axis_dim = 1
    elseif !isnan(yslice) 
        slice_position = yslice
        slice_axis_name = :y
        slice_axis_dim = 2
    elseif !isnan(zslice)
        slice_position = zslice
        slice_axis_name = :z
        slice_axis_dim = 3
    end
    return slice_position, slice_axis_name, slice_axis_dim
end

function get_slice_idx(slice_axis_positions, slice_position)
    slice_idx = findnearest(slice_axis_positions, uconvert(u"mm", slice_position))
    nearest_slice_position = slice_axis_positions[slice_idx]
    return slice_idx, nearest_slice_position
end

findnearest(A::AbstractArray,t) = findmin(abs.(A.-t))[2]

get_slice_data(::Val{:x}, slice_idx, data) = data[slice_idx, :, :]
get_slice_data(::Val{:y}, slice_idx, data) = data[:, slice_idx, :]
get_slice_data(::Val{:z}, slice_idx, data) = data[:, :, slice_idx]

get_slice_image_axes(::Val{:x}) = "yz", [2,3]
get_slice_image_axes(::Val{:y}) = "xz", [1,3]
get_slice_image_axes(::Val{:z}) = "xy", [1,2]
