const PressureArray{x} = Array{T, x} where T <: Union{Unitful.Pressure, Number}
const Volt = typeof(1.0u"V")
const Meter = typeof(1.0u"m")
const Sec = typeof(1.0u"s")


abstract type UltrasoundMetric end

abstract type PNP <: UltrasoundMetric end        # peak negative pressure
abstract type PPP <: UltrasoundMetric end        # peak positive pressure
abstract type ISPPA <: UltrasoundMetric end      # intensity_sppa
abstract type ISPTA <: UltrasoundMetric end      # intensity_spta
abstract type MI <: UltrasoundMetric end         # mechanical_index

struct Metric{T, D}
    val::AbstractArray
end

get_metric_display_name(a::Metric{PNP}) = "Peak negative pressure"
get_metric_display_name(a::Metric{PPP}) = "Peak positive pressure"
get_metric_display_name(a::Metric{ISPPA}) = "Intensity_SPPA"
get_metric_display_name(a::Metric{ISPTA}) = "Intensity_SPTA"
get_metric_display_name(a::Metric{MI}) = "Mechanical Index"

@kwdef struct ScanParameters
    medium::Any
    excitation::Any
    f0::Float64
    hydrophone_id::Symbol
    preamp_id::Union{Symbol, Nothing} = nothing
    calibration_factor::typeof(1.0u"Pa/V") = volt_to_pressure(f0, hydrophone_id, preamp_id)
end

function Base.show(io::IO, scan_setup::ScanParameters)
    println(io, "medium: ", scan_setup.medium)
    println(io, "excitation: ", scan_setup.excitation)
    println(io, "f0: ", scan_setup.f0)
    println(io, "hydrophone_id: ", scan_setup.hydrophone_id)
    println(io, "preamp_id: ", scan_setup.preamp_id)
    println(io, "calibration_factor: ", scan_setup.calibration_factor)
end

struct ScanMetric
    params::ScanParameters
    pressure_trace::PressureArray
    pressure_pn::Metric{PNP}
    pressure_pp::Metric{PPP}
    isppa::Metric{ISPPA}
    isppa_max::Tuple{Unitful.Quantity, Vector{Meter}}
    ispta::Metric{ISPTA}
    ispta_max::Tuple{Unitful.Quantity, Vector{Meter}}
    mechanical_index::Metric{MI}
    mechanical_index_max::Tuple{Float64, Vector{Meter}}
end

"""
Input:
       xyz: The handle of desired xyz stage
     scope: Handle of desired scope
   channel: This input decides which channel the scope will read from.

```
ps = initialize(PositionerSystem)
scope = initialize(Scope350MHz)
hydrophone = IntensityScan(xyz = ps, scope = scope, channel = 1, precapture_delay=0u"s")
```
"""
@kwdef struct IntensityScan 
    xyz::T where T <: ThorlabsLTStage.PositionerSystem
    scope::TcpInstruments.Instr{T} where T <: Oscilloscope
    channel::Int64
    post_move_delay::typeof(1.0u"s") = 0u"s"
    precapture_delay::typeof(1.0u"s")
    trigger_function::Function = () -> nothing
end

function Base.show(io::IO, scan_setup::IntensityScan)
    println(io, "positioning_system: ", scan_setup.xyz)
    println(io, "scope: ", scan_setup.scope)
    println(io, "scope_channel: ", scan_setup.channel)
    println(io, "post_move_delay: ", scan_setup.post_move_delay)
    println(io, "precapture_delay: ", scan_setup.precapture_delay)
    println(io, "trigger_function: ", scan_setup.trigger_function)
end


const version = v"1"


struct ScanData
    version::VersionNumber
    axes::String
    scope_info::TcpInstruments.ScopeInfo
    time::Vector{Sec}
    waveform::Array{Volt, N} where {N}
    coordinates::Array{Meter, N} where {N}
    metrics::Union{Nothing, ScanMetric}
end

function ScanData(
    axes::AbstractString,
    scope_info::TcpInstruments.ScopeInfo,
    time::AbstractArray,
    num_points_per_waveform,
    num_points_per_axis::AbstractArray
)
    waveform_dimensions = Tuple([num_points_per_waveform; num_points_per_axis])
    waveform = u"V" * zeros(waveform_dimensions)
    
    num_coordinates_per_point = 3       # x, y, z location
    coordinates_dimensions = Tuple([num_coordinates_per_point; num_points_per_axis])
    coordinates = u"m" * zeros(coordinates_dimensions)

    return ScanData(version, axes, scope_info, time, waveform, coordinates, nothing)
end


function Base.show(io::IO, scan::ScanData)
    println(io, "version: ", scan.version)
    println(io, "axes: ", scan.axes)
    println(io, "scope_info: ", scan.scope_info)
    println(io, "time: ", size(scan.time))
    println(io, "waveform: ", size(scan.waveform))
    println(io, "coordinates: ", size(scan.coordinates))
    metrics_fieldnames = fieldnames(typeof(scan.metrics))
    if isempty(metrics_fieldnames)
        println(io, "metrics: [none]")
    else
        println(io, "metrics: ")
        for fieldname in metrics_fieldnames
            println(io, "  ", fieldname)
        end
    end
end
