using IntensityScans.IntensityMetrics
using IntensityScans: Metric

@testset "Metrics" begin
    @testset "IntensityMetrics.jl interface" begin
        # ultrasound parameters
        f0 = 5e6u"Hz"                           # center frequency
        pressure_amplitude = 1u"MPa"
        duty_cycle = 1.0
        pulse_duration = 10u"µs"
        total_duration = pulse_duration
        fs = 40e6u"Hz"                          # sampling frequency of received signal
        medium = IntensityMetrics.Medium();
        excitation = IntensityMetrics.Excitation(f0, pulse_duration, duty_cycle, total_duration, fs);

        # waveform parameters
        dt = 1/fs
        num_samples = ceil(pulse_duration/dt)
        time = collect((0:num_samples-1)*dt);
        data = sin.(2*pi*f0*time) * pressure_amplitude;

        @testset "peak negative pressure" begin
            data_all = repeat(data, 1, 3, 2)
            pressure_pn = IntensityScans.calc_pressure_pn(data_all)
            pressure_pn_expected = Metric{IntensityScans.PNP, 2}(repeat([pressure_amplitude], 3, 2))
            @test pressure_pn.val ≈ pressure_pn_expected.val atol=0.001u"Pa"
        end

        @testset "peak positive pressure" begin
            data_all = repeat(data, 1, 3, 2)
            pressure_pp = IntensityScans.calc_pressure_pp(data_all)
            pressure_pp_expected = Metric{IntensityScans.PPP, 2}(repeat([pressure_amplitude], 3, 2))
            @test pressure_pp.val ≈ pressure_pp_expected.val atol=0.001u"Pa"
        end

        @testset "intensity_sppa" begin 
            data_all = repeat(data, 1, 3, 2)
            i_sppa_all_calculated = IntensityScans.calc_intensity_sppa(data_all, medium, excitation)

            i_sppa_expected = uconvert(u"W/cm^2", pressure_amplitude^2 / (2 * medium.density * medium.c))
            i_sppa_all_expected = Metric{IntensityScans.ISPPA, 2}(repeat([i_sppa_expected], 3, 2))

            @test i_sppa_all_calculated.val ≈ i_sppa_all_expected.val atol=0.001u"W/m^2"
        end

        @testset "intensity_spta" begin
            data_all = repeat(data, 1, 3, 2)
            i_spta_all_calculated = IntensityScans.calc_intensity_spta(data_all, medium, excitation)

            i_spta_expected = uconvert(u"W/cm^2", pressure_amplitude^2 / (2 * medium.density * medium.c))
            i_spta_all_expected = Metric{IntensityScans.ISPTA, 2}(repeat([i_spta_expected], 3, 2))

            @test i_spta_all_calculated.val ≈ i_spta_all_expected.val atol=0.001u"W/m^2"
        end

        @testset "mechanical index" begin
            data_all = repeat(data, 1, 3, 2)
            mi_all_calculated = IntensityScans.calc_mechanical_index(data_all, excitation)

            mi_expected = ustrip(uconvert(u"MPa", pressure_amplitude) / sqrt(uconvert(u"MHz", f0)))
            mi_all_expected = Metric{IntensityScans.MI, 2}(repeat([mi_expected], 3, 2))

            @test mi_all_calculated.val ≈ mi_all_expected.val atol=0.001
        end
    end
    

    show(metrics_xyz)

    @test metrics_x.metrics isa ScanMetric
    @test metrics_xy.metrics isa ScanMetric
    @test metrics_xyz.metrics isa ScanMetric

    @test metrics_xyz.metrics.isppa isa Metric{IntensityScans.ISPPA, 3}
    @test metrics_xyz.metrics.ispta isa Metric{IntensityScans.ISPTA, 3}
    @test metrics_xyz.metrics.mechanical_index isa Metric{IntensityScans.MI, 3}

    sppa_max, sppa_xyz = metrics_xyz.metrics.isppa_max
    @test sppa_max isa Unitful.Quantity
    @test length(sppa_xyz) == 3
    @test sppa_xyz isa Vector{typeof(1.0u"m")}

    spta_max, spta_xyz = metrics_xyz.metrics.ispta_max
    @test spta_max isa Unitful.Quantity
    @info spta_max
    @test length(spta_xyz) == 3
    @test spta_xyz isa Vector{typeof(1.0u"m")}

    mi_max, mi_xyz = metrics_xyz.metrics.mechanical_index_max 
    @test mi_max isa Float64
    @test length(mi_xyz) == 3
    @test mi_xyz isa Vector{typeof(1.0u"m")}
end
